/**
 *  Copyright 2009-2011 Matus Chochlik. Distributed under the Boost
 *  Software License, Version 1.0. (See accompanying file
 *  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 *
 *  WARNING: this file is here for development / testing purposes only
 *  and its contents and input / output can change witout any prior
 *  notice.
 */

#include <mirror/mirror_base.hpp>
#include <mirror/pre_registered/type/native.hpp>
#include <mirror/pre_registered/type/std/string.hpp>

#include <puddle/puddle.hpp>

#include <cassert>
#include <iostream>
#include <stdexcept>

//#include "./test.hpp"

PUDDLE_NAMESPACE_BEGIN
namespace type_erased {

template <typename TypeErased>
class range
{
private:
	typedef const TypeErased& (*member_getter)(void);
	const std::vector<member_getter>& getters;
	typename std::vector<member_getter>::const_iterator pos;
public:
	range(const std::vector<member_getter>& _getters)
	 : getters(_getters)
	 , pos(getters.begin())
	{ }

	inline bool empty(void) const
	{
		return getters.end() == pos;
	}

	inline size_t size(void) const
	{
		return getters.end() - pos;
	}

	inline const TypeErased& front(void) const
	{
		assert(!empty());
		return (*pos)();
	}

	inline void step_front(void)
	{
		assert(!empty());
		++pos;
	}
};

template <typename Tag>
struct wrapper_by_tag;

class meta_meta_object;
template <>
struct wrapper_by_tag<mirror::meta_meta_object_tag>
{
	typedef meta_meta_object type;
};

#define PUDDLE_HELPER_MAKE_TAG_TO_META_OBJECT_MAP(OBJECT, I, X) \
class meta_ ## OBJECT; \
template <> \
struct wrapper_by_tag<mirror::meta_ ## OBJECT ## _tag> \
{ \
	typedef meta_ ## OBJECT type; \
}; \
puddle::meta_object_category categorize_wrapper(meta_ ## OBJECT*) \
{ \
	return puddle::meta_object_category::meta_## OBJECT ##_tag; \
}

MIRROR_FOR_EACH_META_OBJECT(PUDDLE_HELPER_MAKE_TAG_TO_META_OBJECT_MAP, _)
#undef PUDDLE_HELPER_MAKE_TAG_TO_META_OBJECT_MAP

template <class MirrorMetaObject>
inline const typename wrapper_by_tag<
	typename mirror::meta_object_category<MirrorMetaObject>::type
>::type& do_wrap(void)
{
	MirrorMetaObject mo;
	static typename wrapper_by_tag<
		typename mirror::meta_object_category<
			MirrorMetaObject
		>::type
	>::type wrapped(mo);
	return wrapped;
}

template <class MirrorMetaObject, class TypeErased>
inline const TypeErased& wrap(void)
{
	return do_wrap<MirrorMetaObject>();
}

template <class TypeErased>
class range_holder
{
private:
	typedef const TypeErased& (*getter)(void);
	std::vector<getter> _getters;

	template <typename ... MetaObjects>
	static std::vector<getter> init_getters(
		mirror::mp::range<MetaObjects...>
	)
	{
		return {
			&wrap<MetaObjects, TypeErased>...
		};
	}
public:
	template <template <class> class Getter, class MetaObject>
	range_holder(Getter<MetaObject>)
	 : _getters(
		init_getters(
			typename mirror::mp::as_range<
				Getter<MetaObject>
			>::type()
		)
	)
	{ }

	range<TypeErased> all(void) const
	{
		return range<TypeErased>(_getters);
	}
};

class specifier
{
private:
	static inline bool category_matches(unsigned a, unsigned b)
	{
		return (a & b) == b;
	}

	specifier_category _cat;
public:
	template <typename Specifier>
	specifier(Specifier s)
	 : _cat(puddle::get_category(s))
	{ }

	friend bool operator == (const specifier& a, const specifier& b)
	{
		return a._cat == b._cat;
	}

	friend bool operator != (const specifier& a, const specifier& b)
	{
		return a._cat != b._cat;
	}

	friend bool operator <= (const specifier& a, const specifier& b)
	{
		return a._cat <= b._cat;
	}

	friend bool operator >= (const specifier& a, const specifier& b)
	{
		return a._cat >= b._cat;
	}

	friend bool operator <  (const specifier& a, const specifier& b)
	{
		return a._cat <  b._cat;
	}

	friend bool operator >  (const specifier& a, const specifier& b)
	{
		return a._cat >  b._cat;
	}

	specifier_category category(void) const
	{
		return _cat;
	}

	inline bool is_a(specifier_category cat) const
	{
		return category_matches(
			unsigned(category()),
			unsigned(cat)
		);
	}

#define PUDDLE_HELPER_MAKE_SPEC_IS_X(SPEC, KW, GRP, I, X) \
	inline bool is_## SPEC(void) const \
	{ \
		return is_a(specifier_category::spec_ ## SPEC ## _tag); \
	}
MIRROR_FOR_EACH_SPECIFIER(PUDDLE_HELPER_MAKE_SPEC_IS_X, _)
#undef PUDDLE_HELPER_MAKE_SPEC_IS_X

};

class meta_meta_object
{
private:
	std::string(*_get_base_name)(void);
	std::string(*_get_construct_name)(void);

	static inline bool category_matches(unsigned a, unsigned b)
	{
		return (a & b) == b;
	}

	meta_object_category _cat;
public:
	template <typename MetaMetaObject>
	meta_meta_object(MetaMetaObject mo)
	 : _get_base_name(&MetaMetaObject::base_name)
	 , _get_construct_name(&MetaMetaObject::construct_name)
	 , _cat(puddle::get_category(typename MetaMetaObject::category()))
	{ }

	inline std::string base_name(void) const
	{
		assert(_get_base_name != nullptr);
		return _get_base_name();
	}

	inline std::string construct_name(void) const
	{
		assert(_get_construct_name != nullptr);
		return _get_construct_name();
	}

	meta_object_category category(void) const
	{
		return _cat;
	}

	// Returns true if the reflected meta-object is of the cat category
	inline bool is_a(meta_object_category cat) const
	{
		return category_matches(
			unsigned(category()),
			unsigned(cat)
		);
	}

#define PUDDLE_HELPER_MAKE_MO_IS_X(OBJECT, I, X) \
	inline bool is_meta_## OBJECT(void) const \
	{ \
		return is_a(meta_object_category::meta_ ## OBJECT ## _tag); \
	}
MIRROR_FOR_EACH_META_OBJECT(PUDDLE_HELPER_MAKE_MO_IS_X, _)
#undef PUDDLE_HELPER_MAKE_MO_IS_X
};

class meta_object
{
private:
	struct uid_holder
	{
	private:
		static int& _counter(void)
		{
			static int val = 0;
			return val;
		}

		template <typename MetaObject>
		static int _init(MetaObject)
		{
			static int val = ++_counter();
			return val;
		}
		int value;
	public:
		template <typename MetaObject>
		uid_holder(MetaObject mo)
		 : value(_init(mo))
		{ }

		int get(void) const
		{
			return value;
		}
	} _uid;

	const meta_meta_object& (*_get_self)(void);

public:
	template <typename MetaObject>
	meta_object(MetaObject mo)
	 : _uid(mo)
	 , _get_self(
		&wrap<MIRRORED_META_OBJECT(MetaObject), meta_meta_object>
	)
	{ }

	inline int uid(void) const
	{
		return _uid.get();
	}

	friend bool operator == (const meta_object& a, const meta_object& b)
	{
		return a.uid() == b.uid();
	}

	friend bool operator != (const meta_object& a, const meta_object& b)
	{
		return a.uid() != b.uid();
	}

	friend bool operator <= (const meta_object& a, const meta_object& b)
	{
		return a.uid() <= b.uid();
	}

	friend bool operator >= (const meta_object& a, const meta_object& b)
	{
		return a.uid() >= b.uid();
	}

	friend bool operator <  (const meta_object& a, const meta_object& b)
	{
		return a.uid() <  b.uid();
	}

	friend bool operator >  (const meta_object& a, const meta_object& b)
	{
		return a.uid() >  b.uid();
	}

	const meta_meta_object& self(void) const
	{
		assert(_get_self != nullptr);
		return _get_self();
	}

	meta_object_category category(void) const
	{
		return self().category();
	}

#define PUDDLE_HELPER_MAKE_MO_IS_X(OBJECT, I, X) \
	inline bool is_## OBJECT(void) const \
	{ \
		return self().is_meta_## OBJECT(); \
	}
MIRROR_FOR_EACH_META_OBJECT(PUDDLE_HELPER_MAKE_MO_IS_X, _)
#undef PUDDLE_HELPER_MAKE_MO_IS_X

/*
	template <typename Interface>
	const Interface& as_a(void) const
	{
		//TODO
	}
*/
};

class meta_named_object
 : virtual public meta_object
{
private:
	std::string(*_get_base_name)(void);
public:
	template <typename MetaNamedObject>
	meta_named_object(MetaNamedObject mo)
	 : meta_object(mo)
	 , _get_base_name(&MetaNamedObject::base_name)
	{ }

	inline std::string base_name(void) const
	{
		assert(_get_base_name != nullptr);
		return _get_base_name();
	}
};

class meta_scoped_object
 : virtual public meta_object
{
private:
	const meta_scope& (*_get_scope)(void);
public:
	template <typename MetaScopedObject>
	meta_scoped_object(MetaScopedObject mo)
	 : meta_object(mo)
	 , _get_scope(&wrap<typename MetaScopedObject::scope, meta_scope>)
	{ }

	inline const meta_scope& scope(void) const
	{
		assert(_get_scope != nullptr);
		return _get_scope();
	}
};

class meta_named_scoped_object
 : public meta_scoped_object
 , public meta_named_object
{
private:
	std::string(*_get_full_name)(void);
	std::string(*_get_local_name)(void);
public:
	template <typename MetaNamedScopedObject>
	meta_named_scoped_object(MetaNamedScopedObject mo)
	 : meta_object(mo)
	 , meta_scoped_object(mo)
	 , meta_named_object(mo)
	 , _get_full_name(&MetaNamedScopedObject::full_name)
	 , _get_local_name(&MetaNamedScopedObject::local_name)
	{ }

	inline std::string full_name(void) const
	{
		assert(_get_full_name != nullptr);
		return _get_full_name();
	}

	inline std::string local_name(void) const
	{
		assert(_get_local_name != nullptr);
		return _get_local_name();
	}
};

class meta_scope
 : virtual public meta_named_scoped_object
{
private:
	range_holder<meta_named_scoped_object> _members;
public:
	template <typename MetaScope>
	meta_scope(MetaScope mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , _members(mirror::members<MetaScope>())
	{ }

	range<meta_named_scoped_object> members(void) const
	{
		return _members.all();
	}
};

class meta_namespace
 : public meta_scope
{
private:
public:
	template <typename MetaNamespace>
	meta_namespace(MetaNamespace mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , meta_scope(mo)
	{ }
};

class meta_global_scope
 : public meta_namespace
{
private:
public:
	template <typename MetaGlobalScope>
	meta_global_scope(MetaGlobalScope mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , meta_namespace(mo)
	{ }
};

class meta_type
 : virtual public meta_named_scoped_object
{
private:

public:
	template <typename MetaType>
	meta_type(MetaType mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	{ }
};

class meta_typedef
 : public meta_type
{
private:
	const meta_type& (*_get_type)(void);
public:
	template <typename MetaTypedef>
	meta_typedef(MetaTypedef mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , meta_type(mo)
	 , _get_type(&wrap<typename MetaTypedef::type, meta_type>)
	{ }

	const meta_type& type(void) const
	{
		assert(_get_type != nullptr);
		return _get_type();
	}
};

class meta_type_template
 : public meta_named_scoped_object
{
private:
public:
	template <typename MetaTypeTemplate>
	meta_type_template(MetaTypeTemplate mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	{ }
};

class meta_templated_type
 : public meta_type
{
private:
	const meta_type_template& (*_get_template)(void);
public:
	template <typename MetaTemplatedType>
	meta_templated_type(MetaTemplatedType mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , meta_type(mo)
	 , _get_template(
		&wrap<
			typename MetaTemplatedType::type_template,
			meta_type_template
		>
	)
	{ }
};

class meta_class
 : public meta_type
 , public meta_scope
{
private:
	specifier _elaborated_type;
public:
	template <typename MetaClass>
	meta_class(MetaClass mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , meta_type(mo)
	 , meta_scope(mo)
	 , _elaborated_type(typename MetaClass::elaborated_type())
	{ }

	const specifier& elaborated_type(void) const
	{
		return _elaborated_type;
	}
};

class meta_class_or_enum_member
 : virtual public meta_named_scoped_object
{
private:
	int _position;
public:
	template <typename MetaClassOrEnumMember>
	meta_class_or_enum_member(MetaClassOrEnumMember mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , _position(MetaClassOrEnumMember::position::value)
	{ }

	inline int position(void) const
	{
		return _position;
	}
};

class meta_class_member
 : virtual public meta_class_or_enum_member
{
private:
	specifier _access_type;
public:
	template <typename MetaClassMember>
	meta_class_member(MetaClassMember mo)
	 : meta_object(mo)
	 , meta_named_scoped_object(mo)
	 , meta_class_or_enum_member(mo)
	 , _access_type(typename MetaClassMember::access_type())
	{ }

	const specifier& access_type(void) const
	{
		return _access_type;
	}

};

} // namespace type_erased

template <class MetaObject>
type_erased::meta_type erase_type(const aux::meta_type<MetaObject>& mo)
{
	return type_erased::meta_type(MetaObject());
}

PUDDLE_NAMESPACE_END

void print_scope(const puddle::type_erased::meta_scope& scope)
{
	auto r = scope.members();
	while(!r.empty())
	{
		std::cout <<
			r.front().self().construct_name() <<
			" " <<
			r.front().base_name() <<
			std::endl;
		//if(r.front().is_scope())
			//print_scope(r.front().as_a<puddle::type_erased::meta_scope>());
		r.step_front();
	}
}

int main(void)
{
	try
	{
		auto x = puddle::adapt<mirror::reflected<unsigned>::type>();
		auto e = puddle::erase_type(x);
		std::cout << e.base_name() << std::endl;
		std::cout << (e.is_type() ? "true" : "false") << std::endl;
		std::cout << (e.scope().is_type() ? "true" : "false") << std::endl;
		std::cout << e.scope().members().size() << std::endl;
		//
		print_scope(e.scope());
		std::cout << std::endl;
	}
	catch(std::exception const& error)
	{
		std::cerr << "Error: " << error.what() << std::endl;
	}
	//
	return 0;
}
